
local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
local plugin = RW:Plugin("Options")

----------------------------
-- Locals

local ACD = LibStub("AceConfigDialog-3.0")
local LoD = RW:GetPlugin("LoD")
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("RW-2.0")

local iconCoords = {
	STAR		= {0, 		0.25,	0, 		0.25},
	CIRCLE 		= {0.25,	0.5,	0,		0.25},
	DIAMOND		= {0.5, 	0.75,	0,		0.25},
	TRIANGLE 	= {0.75,	1,		0,		0.25},
	MOON 		= {0,		0.25,	0.25,	0.5},
	SQUARE 		= {0.25,	0.5,	0.25,	0.5},
	CROSS 		= {0.5,		0.75,	0.25,	0.5},
	SKULL 		= {0.75,	1,		0.25,	0.5},
}

local OptionsFrame
local DetailsFrame
local bosses
local bossMods

----------------------------
-- Init

function plugin:OnRegister()
	
	self.BossOptions = {BossMods = {}}
	self.PluginOptions = {Plugins = {}}
	
	self:RegisterMessage("RW_BossModFound")
	self:RegisterMessage("RW_BossRegistered")
	
	self.options = {
		type = "group",
		name = "Raid Watch 2",
		handler = self,
		args = {
			General = {
				type = "group",
				name = "General",
				order = 1,
				args = {
					Config = {
						type = "execute",
						name = "Config",
						func = "ToggleConfigMode",
					},
				}
			},
			Plugins = {	
				type = "group",
				name = "Plugins",
				order = 3,
				plugins = self.PluginOptions,
				args = {}
			}
		}
	}	
	
	bosses = {}
	bossMods = {}
end

function plugin:PluginEnable()
	self.NextConfigMode = true
	LibStub("AceConfig-3.0"):RegisterOptionsTable("RW", self.options)
	
	self:RegisterMessage("RW_BossRegistered")
	self:RegisterMessage("RW_BossModFound")
end

function plugin:PluginDisable()
end

function plugin:ToggleConfigMode()
	RW:SendMessage("RW_ConfigMode", self.NextConfigMode) 
	self.NextConfigMode = not self.NextConfigMode
end

----------------------
-- Message handlers --
----------------------
function plugin:RW_BossRegistered(message, boss)
	bosses[boss.modPack] = bosses[boss.modPack] or {}
	tinsert(bosses[boss.modPack], boss)
end

function plugin:RW_BossModFound(message, modName, modID)
	bossMods[modName] = modID or true
end
----------------------

do
	local treeSelectPath = {}
	local treeStatusTable = {}
	local treeLastSelected	
	
	local bossList = {}
	local bossListLastSelected = {}
	local bossHeaders = {}
	
	------------------------------
	-- Option Details
	
	local function showDetailTab(tabs, event, group)
		tabs:ReleaseChildren()
		local option, boss = tabs:GetUserData("option"), tabs:GetUserData("boss")
		
		local scroll = AceGUI:Create("ScrollFrame")
		scroll:SetLayout("Flow")
		scroll:SetFullWidth(true)
		scroll:SetFullHeight(true)
		
		
		scroll:AddChild(boss.optionGets[group](boss, option))
		
		tabs:AddChild(scroll)
	end
	
	-- Returns the tabs for the details about the selected option
	local tabs = {}
	local function getOptionTabs(option, boss)
		wipe(tabs)
		for i, b in pairs(RW.BITS) do
			if bit.band(option.uses, b) == b then
				tinsert(tabs, {value = i, text = L[i]})
			end
		end
		return tabs
	end
	
	local function showHideDetails(option, boss)
		if DetailsFrame then 
			DetailsFrame:Release()
			DetailsFrame = nil
		end
			
		DetailsFrame = AceGUI:Create("RWFrame")
		DetailsFrame:SetLayout("Fill")
		DetailsFrame.frame:ClearAllPoints()
		DetailsFrame.frame:SetPoint("LEFT", OptionsFrame.frame, "RIGHT", 5, 0)
		DetailsFrame:SetWidth(500)
		DetailsFrame:SetHeight(450)
		DetailsFrame:SetTitle(option.text)
		DetailsFrame:SetCallback("OnClose", function(widget)
			widget:ReleaseChildren()
			widget:Release()
			DetailsFrame = nil
		end)
		
		local t = getOptionTabs(option, boss)
		local tab = AceGUI:Create("TabGroup")
		tab:SetTabs(t)
		tab:SetUserData("option", option)
		tab:SetUserData("boss", boss)
		tab:SetCallback("OnGroupSelected", showDetailTab)
		tab:SelectTab(t[1] and t[1].value or nil)
		tab:SetCallback("OnTabEnter", function(widget, event, key, tab)
				GameTooltip:SetOwner(widget.frame)
				GameTooltip:AddLine(L[key])
				GameTooltip:AddLine(L[key.."_DESC"])
				GameTooltip:Show()
			end)
		tab:SetCallback("OnTabLeave", function() GameTooltip:Hide() end)
		
		DetailsFrame:AddChild(tab)
		
		DetailsFrame:Show()
	end
	
	------------------------------
	-- Boss Detials
	
	local function getMasterValue(option, boss)
		if boss.db.profile[option.text].uses == 0 then
			return false
		end
		if boss.db.profile[option.text].uses == boss.defaultOptions[option.text].uses then
			return true 
		end
		return nil
	end
	
	local function toggleMaster(check, event, value)
		local parent = check:GetUserData("parent")
		local boss, option = parent:GetUserData("boss"), parent:GetUserData("option")
		if value == nil then check:SetValue(false) end
		if value then
			boss.db.profile[option.text].uses = boss.defaultOptions[option.text].uses
		else
			boss.db.profile[option.text].uses = 0
		end		
		if DetailsFrame then
			showHideDetails(option, boss)
		end
	end
	
	local function toggleBool(check, event, value)
		local parent = check:GetUserData("parent")
		local boss, option = parent:GetUserData("boss"), parent:GetUserData("option")
		boss.db.profile[option.optionName] = value
	end
	
	local function getBoolOption(option, boss)
		local group = AceGUI:Create("OptionFrame")
		group:SetLayout("Flow")
		group:SetTitle(option.text)
		group:SetRelativeWidth(0.5)
		group:SetAutoAdjustHeight(false)
		group:SetHeight(80)
		group:SetUserData("boss", boss)
		group:SetUserData("option", option)
		
		local check = AceGUI:Create("CheckBox")
		check:SetLabel("Enable")
		check:SetUserData("parent", group)
		check:SetValue(boss.db.profile[option.optionName])
		check:SetCallback("OnValueChanged", toggleBool)
		
		group:AddChild(check)
		
		return group
	end
	
	local function getGeneralOption(option, boss, widget)
		local group = AceGUI:Create("OptionFrame")
		group:SetLayout("Flow")
		group:SetTitle(option.text)
		group:SetRelativeWidth(0.5)
		group:SetAutoAdjustHeight(false)
		group:SetHeight(80)
		group:SetUserData("boss", boss)
		group:SetUserData("option", option)
		
		local button = AceGUI:Create("Button")
		button:SetWidth(80)
		button:SetHeight(40)
		button:SetText("Details")
		button:SetCallback("OnClick", function(self, event, button)
			showHideDetails(self.parent:GetUserData("option"), self.parent:GetUserData("boss"))
		end)
		
		local check = AceGUI:Create("CheckBox")
		check:SetTriState(true)
		check:SetLabel("Enable")
		check:SetUserData("parent", group)
		check:SetRelativeWidth(0.5)
		check:SetValue(getMasterValue(option, boss))
		check:SetCallback("OnValueChanged", toggleMaster)
		
		group:AddChildren(button)
		group:AddChildren(check)
		
		return group
	end
	
	local function getSpellOption(option, boss, widget)
		local group = AceGUI:Create("OptionFrame")
		group:SetLayout("Flow")
		group:SetTitle(option.text)
		group:SetRelativeWidth(0.5)
		group:SetAutoAdjustHeight(false)
		group:SetHeight(80)
		group:SetUserData("boss", boss)
		group:SetUserData("option", option)
		
		local icon = AceGUI:Create("Icon")
		icon:SetText("")
		icon:SetImage(option.icon, 0, 1, 0, 1)
		icon:SetImageSize(35, 35)
		icon:SetHeight(40)
		icon:SetWidth(50)
		--icon:SetRelativeWidth(0.4)
		icon:SetCallback("OnEnter", function(self)
			GameTooltip:SetOwner(self.frame)
			GameTooltip:SetHyperlink("spell:"..self.parent:GetUserData("option").spell)
			GameTooltip:AddLine(" ")
			GameTooltip:AddLine("Click to show detailed settings")
			GameTooltip:AddLine("[Shift + Left click to add link in chat]")
			GameTooltip:Show()
		end)
		icon:SetCallback("OnLeave", function()
			GameTooltip:Hide()
		end)
		icon:SetCallback("OnClick", function(self, event, button)
			if button == "LeftButton" then
				if IsShiftKeyDown() and ChatFrameEditBox:IsShown() then
					ChatFrameEditBox:Insert("\124cff71d5ff\124Hspell:"..self.parent:GetUserData("option").spell.."\124h["..self.parent:GetUserData("option").text.."]\124h\124r")
				else
					showHideDetails(self.parent:GetUserData("option"), self.parent:GetUserData("boss"))
				end
			end
		end)
		
		local check = AceGUI:Create("CheckBox")
		check:SetTriState(true)
		check:SetLabel("Enable")
		check:SetUserData("parent", group)
		check:SetRelativeWidth(0.5)
		check:SetValue(getMasterValue(option, boss))
		check:SetCallback("OnValueChanged", toggleMaster)
		
		group:AddChildren(icon, check)
		
		return group
	end
	
	local function selectBoss(drop, event, group)
		drop:ReleaseChildren()
		
		boss = bosses[drop:GetUserData("bossMod")][group]
		if not boss then error(("Bossmod %q is not loaded"):format(drop:GetUserData("bossMod"))) end
		drop:SetUserData("boss", boss)
		
		bossListLastSelected[drop:GetUserData("bossMod")] = group
		
		local scroll = AceGUI:Create("ScrollFrame")
		scroll:SetFullWidth(true)
		scroll:SetFullHeight(true)
		scroll:SetLayout("Flow")

		local miscGroup = AceGUI:Create("CategoryFrame")
		miscGroup:SetLayout("Flow")
		miscGroup:SetTitle("Misc")
		miscGroup:SetRelativeWidth(0.5)
		scroll:AddChild(miscGroup)
		
		local enableCheck = AceGUI:Create("CheckBox")
		enableCheck:SetLabel("Enabled")
		enableCheck:SetHeight(35)
		enableCheck:SetValue(boss.db.profile.enabled)
		miscGroup:AddChild(enableCheck)
		
		if boss.usedIcons then
			local iconGroup = AceGUI:Create("CategoryFrame")
			iconGroup:SetLayout("Flow")
			iconGroup:SetTitle("Icons")
			iconGroup:SetRelativeWidth(0.5)
			scroll:AddChild(iconGroup)
			
			for i in pairs(boss.usedIcons) do
				local icon = AceGUI:Create("Icon")
				icon:SetImage("Interface\\TargetingFrame\\UI-RaidTargetingIcons.blp", unpack(iconCoords[i]))
				icon:SetImageSize(20, 20)
				icon:SetRelativeWidth(0.124)
				icon:SetHeight(35)
				iconGroup:AddChild(icon)
			end
		end
		
		if boss.options then
			-- Build options titles
			wipe(bossHeaders)
			if boss.options.headers then
				for i, header in pairs(boss.options.headers) do
					local group = AceGUI:Create("CategoryFrame")
					group:SetLayout("Flow")
					group:SetTitle(header)
					group:SetFullWidth(true)
					scroll:AddChild(group)
					bossHeaders[header] = group
				end
			end
			-- Add all the options
			for i, option in pairs(boss.options) do
				if option.type == "spell" then
					if option.header and bossHeaders[option.header] then
						bossHeaders[option.header]:AddChildren(getSpellOption(option, boss))
					end
				elseif option.type == "bool" then
					if option.header and bossHeaders[option.header] then
						bossHeaders[option.header]:AddChildren(getBoolOption(option, boss))
					end
				--[[
				elseif option.type == "slider" then
					if option.header and bossHeaders[option.header] then
						bossHeaders[option.header]:AddChildren(getSliderOption(option, boss))
					end ]]--
				elseif option.type == "general" then
					if option.header and bossHeaders[option.header] then
						bossHeaders[option.header]:AddChildren(getGeneralOption(option, boss))
					end
				end 
			end
		end		
		
		drop:AddChild(scroll)
	end
	
	------------------------------
	-- Boss Selection
	
	local function buildBossList(modName, subGroup)
		local firstNr = nil
		wipe(bossList)
		for	i, boss in pairs(bosses[modName]) do
			if boss.subGroup == subGroup then
				if not firstNr then firstNr = i end
				bossList[i] = boss:GetName()
			end
		end
		return bossList, firstNr
	end
	
	local function selectTreeOption(tree, event, group)
		tree:ReleaseChildren()
		treeLastSelected = group
		
		-- Only make AceGUI configuration if its a boss mod
		-- Else open host the AceConfig inside this window
		if select(1, string.split("\001", group)) ~= "BossMods" then
			wipe(treeSelectPath)
			for i = 1, select("#", string.split("\001", group)) do
				pathPart = select(i, string.split("\001", group))
				if not pathPart:find("category") then 
					tinsert(treeSelectPath, pathPart)
				end
			end
			local bgroup = AceGUI:Create("SimpleGroup")
			bgroup:SetLayout("Fill")
			bgroup:SetFullWidth(true)
			bgroup:SetFullHeight(true)
			tree:AddChild(bgroup)
			if #treeSelectPath == 1 and treeSelectPath[1] == "Plugins" then return end
			ACD:Open("RW", bgroup, unpack(treeSelectPath))
		else
			-- Get the full path minus "BossMods"
			nodes = { select(2, string.split("\001", group)) }
			if #nodes < 1 then return end
			
			local bossMod, subGroup
			-- This mod has sub groups
			if #nodes > 1 then
				bossMod = select(4, nodes[#nodes - 1]:find("type:([%w%s]*);?([%w%s]*)"))
				subGroup = select(4, nodes[#nodes]:find("type:([%w%s]*);?([%w%s]*)"))
			-- This mod have no subgroups
			else
				bossMod = select(4, nodes[#nodes]:find("type:([%w%s]*);?([%w%s]*)"))
			end
			-- This mod is loaded
			if bosses[bossMod] then
				nameList, firstNr = buildBossList(bossMod, subGroup)
				if firstNr then
					dropdown = AceGUI:Create("DropdownGroup")
					dropdown:SetLayout("Flow")
					dropdown:SetUserData("bossMod", bossMod)
					dropdown:SetFullWidth(true)
					dropdown:SetFullHeight(true)
					dropdown:SetGroupList(nameList)
					dropdown:SetCallback("OnGroupSelected", selectBoss)
					dropdown:SetGroup(bossListLastSelected[bossMod] or firstNr)
					tree:AddChild(dropdown)
				end
			-- Not loaded, (add load button)
			else
				local button = AceGUI:Create("Button")
				button:SetText("Load")
				button:SetUserData("modID", bossMods[bossMod])
				button:SetUserData("modName", bossMod)
				button:SetUserData("tree", tree)
				button:SetUserData("path", group)
				button:SetCallback("OnClick", function(widget, event)
					if not IsAddOnLoaded(widget:GetUserData("modID")) then
						LoD:LoadMod(widget:GetUserData("modID"), widget:GetUserData("modName"))
						widget:GetUserData("tree"):SetTree(plugin:BuildOptionsTree())
						selectTreeOption(widget:GetUserData("tree"), "nil", widget:GetUserData("path"))
					end
				end)
				tree:AddChild(button)
			end
		end
	end

	
	-----------------------------
	-- Main window
	
	function plugin:Toggle()
		if OptionsFrame then
			OptionsFrame:Release()
			OptionsFrame = nil
		else		
			OptionsFrame = AceGUI:Create("RWFrame")
			OptionsFrame:SetTitle("Raid Watch 2")
			OptionsFrame:SetLayout("Fill")
			OptionsFrame:SetWidth(810)
			OptionsFrame:SetHeight(600)
			OptionsFrame:SetCallback("OnClose", function(widget)
				widget:ReleaseChildren()
				widget:Release()
				if DetailsFrame then
					DetailsFrame:Release()
					DetailsFrame = nil
				end
				OptionsFrame = nil
			end)
			OptionsFrame.frame:ClearAllPoints()
			OptionsFrame.frame:SetPoint("LEFT", UIParent, "LEFT", 50, 0)
			
			local tree = AceGUI:Create("TreeGroup")
			tree:SetLayout("Flow")
			tree:SetTree(self:BuildOptionsTree())
			tree:SetCallback("OnGroupSelected", selectTreeOption)		
			tree:SetSelected(treeLastSelected or "General")	
			tree:SetStatusTable(treeStatusTable)
			OptionsFrame:AddChild(tree)
			
			OptionsFrame:Show()
		end
	end
end

do
	local catOrders = {}
	local TreeGroups = {}
	function plugin:BuildOptionsTree()
		wipe(TreeGroups)
		TreeGroups = {
			{
				value = "General",
				text = "General"
			},
			{
				value = "BossMods",
				text = "Boss Mods",
				children = {
				}
			},
			{
				value = "Plugins",
				text = "Plugins",
				children = {
				}
			}
		}
		for i, module in RW:IteratePlugins() do
			if module.GetOptions then
				local pluginOptions = module:GetOptions()
				self.PluginOptions.Plugins[module:GetName()] = pluginOptions
				-- Module has a category
				if module.category then
					-- First time category seen this round
					if not catOrders[module.category] then
						TreeGroups[3].children[#TreeGroups[3].children + 1] = {value = "category_"..module.category, text = module.category, children = {}, icon = [[Interface\ICONS\Ability_Creature_Cursed_02]]}
						catOrders[module.category] = #TreeGroups[3].children
					-- Category seen before but not added yet
					elseif not TreeGroups[3].children[catOrders[module.category]] then
						TreeGroups[3].children[catOrders[module.category]] = {value = "category_"..module.category, text = module.category, children = {}, icon = [[Interface\ICONS\Ability_Creature_Cursed_02]]}
					end
					-- Add module to the category
					tinsert(TreeGroups[3].children[catOrders[module.category]].children, {value = module:GetName(), text = pluginOptions.name})			
				else
					tinsert(TreeGroups[3].children, {value = module:GetName(), text = pluginOptions.name})
				end
			end
		end
		
		for i, cat in pairs(TreeGroups[3].children) do
			if cat.children then
				table.sort(cat.children, function(a, b) return a.text < b.text end)
			end
		end
		
		local function checkList(lst, value)
			for i, t in pairs(lst) do
				if t.text == value then return false end
			end
			return true
		end
		
		if RW:NrElements(bossMods) == 0 then
			RW:GetModsAndBosses()
		end
		for modName, modID in pairs(bossMods) do
			leafe = {value = "type:leafe;"..modName, text = modName}
			if bosses[modName] then
				for i, boss in pairs(bosses[modName]) do
					if boss.subGroup then
						if not leafe.children or checkList(leafe.children, boss.subGroup) then
							leafe.value = "type:subGroup;"..modName
							leafe.children = leafe.children or {}
							tinsert(leafe.children, {value = "type:leafe;"..boss.subGroup, text = boss.subGroup})
						end
					end
				end
			end		
			tinsert(TreeGroups[2].children, leafe)
		end
				
		return TreeGroups
	end
end
